import asyncio
import json
import logging
import os
import re
import tempfile
import traceback
import typing
from pathlib import Path

import numpy
import pyaudio
from pydub import AudioSegment
from pydub.silence import detect_leading_silence
from pydub.utils import make_chunks, ratio_to_db
from yapper.utils import get_random_name

from py_speech_service import speech_service_pb2
from py_speech_service.piper import Piper

trim_leading_silence = lambda x: x[detect_leading_silence(x):]
trim_trailing_silence = lambda x: trim_leading_silence(x.reverse()).reverse()
strip_silence = lambda x: trim_trailing_silence(trim_leading_silence(x))

class SpeechSettings:
    model_name: str = "hfc_female"
    onnx_path: str = ""
    config_path: str = ""
    alt_model_name: str = "hfc_male"
    alt_onnx_path: str = ""
    alt_config_path: str = ""
    speed: float = 1
    pitch: float = 1
    gain: float = 0
    
    def __init__(self, other = None):
        if type(other) is SpeechSettings:
            self.model_name = other.model_name
            self.onnx_path = other.onnx_path
            self.config_path = other.config_path
            self.alt_model_name = other.alt_model_name
            self.alt_onnx_path = other.alt_onnx_path
            self.alt_config_path = other.alt_config_path
            self.speed = other.speed
            self.pitch = other.pitch
            self.gain = other.gain
        elif other is not None:
            self.model_name = other.model_name if hasattr(other, "model_name") and other.model_name else "hfc_female"
            self.onnx_path = other.onnx_path if hasattr(other, "onnx_path") and other.onnx_path else ""
            self.config_path = other.config_path if hasattr(other, "config_path") and other.config_path else ""
            self.alt_model_name = other.alt_model_name if hasattr(other, "alt_model_name") and other.alt_model_name else "hfc_male"
            self.alt_onnx_path = other.alt_onnx_path if hasattr(other, "alt_onnx_path") and other.alt_onnx_path else ""
            self.alt_config_path = other.alt_config_path if hasattr(other, "alt_config_path") and other.alt_config_path else ""
            self.speed = numpy.clip(other.speed, 0.5, 2.0) if hasattr(other, "speed") and other.speed else 1
            self.gain = numpy.clip(other.gain, -100.0, 100.0) if hasattr(other, "gain") and other.gain else 0
            self.pitch = numpy.clip(other.pitch, 0.5, 1.5) if hasattr(other, "pitch") and other.pitch else 1


class PendingSpeechRequest:
    message: str
    silence_seconds: float
    speech_settings: SpeechSettings
    use_alt_voice: bool

    original_message: str
    first_request_of_message: bool
    last_request_of_message: bool
    file_path: str
    message_id: int

    def __init__(self):
        self.file_path = os.path.join(tempfile.gettempdir(), "py_speech_service", get_random_name(20) + ".wav")

    def to_string(self):
        data = {
            "message": str(self.message) if hasattr(self, "message") else "",
            "silence_seconds": str(self.silence_seconds) if hasattr(self, "silence_seconds") else "",
            "model_name": str(self.speech_settings.model_name) if hasattr(self, "model_name") else "",
            "onnx_path": str(self.speech_settings.onnx_path) if hasattr(self, "onnx_path") else "",
            "config_path": str(self.speech_settings.config_path) if hasattr(self, "config_path") else "",
            "use_alt_voice": str(self.use_alt_voice) if hasattr(self, "use_alt_voice") else "",
            "speed_modifier": str(self.speech_settings.speed) if hasattr(self, "speech_settings") else "",
            "gain_modifier": str(self.speech_settings.gain) if hasattr(self, "speech_settings") else "",
            "pitch_modifier": str(self.speech_settings.pitch) if hasattr(self, "speech_settings") else "",
        }
        return json.dumps(data)

class Speaker:

    process_queue: asyncio.Queue[PendingSpeechRequest] = asyncio.Queue()
    play_queue: asyncio.Queue[PendingSpeechRequest] = asyncio.Queue()
    grpc_response_queue: typing.Optional[asyncio.Queue] = None
    shutdown_event = asyncio.Event()
    stop_talking_event = asyncio.Event()

    speech_settings: SpeechSettings = SpeechSettings()
    piper: typing.Optional[Piper] = None
    is_done = False
    is_speaking = False
    volume: float = 1

    def start(self):
        folder = os.path.join(tempfile.gettempdir(), "py_speech_service")
        if not Path(folder).exists():
            Path(folder).mkdir()
        asyncio.create_task(self.handle_process_queue())
        asyncio.create_task(self.handle_play_queue())

    async def speak_basic_line(self, line: str):
        folder = os.path.join(tempfile.gettempdir(), "py_speech_service")
        if not Path(folder).exists():
            Path(folder).mkdir()
        asyncio.create_task(self.handle_process_queue())
        asyncio.create_task(self.handle_play_queue())
        response_queue = asyncio.Queue()
        self.set_grpc_response_queue(response_queue)
        await self.speak(line)
        while not self.shutdown_event.is_set():  # Keep running unless shutdown is triggered
            try:
                response = await asyncio.wait_for(response_queue.get(), timeout=1.0)  # Wait for 1 second
                if hasattr(response.speak_update, "is_end_of_message") and response.speak_update.is_end_of_message:
                    break
            except asyncio.TimeoutError:
                continue  # Retry checking

    def init_speech_settings(self, settings: SpeechSettings):
        if self.piper is None:
            self.piper = Piper()
        self.piper.set_speech_settings(settings.alt_onnx_path, settings.alt_config_path, settings.alt_model_name)
        self.piper.set_speech_settings(settings.onnx_path, settings.config_path, settings.model_name)
        self.speech_settings = settings
        response = speech_service_pb2.SpeechServiceResponse()
        response.speech_settings_set.successful = True
        if self.grpc_response_queue:
            asyncio.create_task(self.grpc_response_queue.put(response))

    async def handle_process_queue(self):
        logging.info("Started handling process queue")
        print("Started handling process queue")
        while not self.shutdown_event.is_set():  # Keep running unless shutdown is triggered
            try:
                item = await asyncio.wait_for(self.process_queue.get(), timeout=.25)  # Wait for 1 second
                await asyncio.create_task(self.__handle_request(item))
            except asyncio.TimeoutError:
                continue  # Retry checking
        logging.info("Stopped handling process queue")
        print("Stopped handling process queue")
        while not self.process_queue.empty():
            self.process_queue.get_nowait()
            self.process_queue.task_done()

    async def handle_play_queue(self):
        logging.info("Started handling play queue")
        print("Started handling play queue")
        while not self.shutdown_event.is_set():  # Keep running unless shutdown is triggered
            try:
                item = await asyncio.wait_for(self.play_queue.get(), timeout=.25)
                await asyncio.create_task(self.__handle_play(item))
            except asyncio.TimeoutError:
                continue  # Retry checking
        logging.info("Stopped handling play queue")
        print("Stopped handling play queue")
        while not self.play_queue.empty():
            self.play_queue.get_nowait()
            self.play_queue.task_done()

    def set_grpc_response_queue(self, queue: asyncio.Queue):
        self.grpc_response_queue = queue

    def split_into_lines(self, paragraph: str, words_per_line: int) -> list[str]:
        words = paragraph.split()  # Split by spaces
        return [' '.join(words[i:i + words_per_line]) for i in range(0, len(words), words_per_line)]

    async def speak(self, message: str, settings: typing.Optional[SpeechSettings] = None, message_id: int = 0):

        self.stop_talking_event.clear()

        if not message.__contains__("</") and not message.__contains__("/>"):
            lines = re.split(r'(?<=[.!?])\s+|\n+', message)
            for line in lines:
                request = PendingSpeechRequest()
                if settings is None:
                    request.speech_settings = SpeechSettings(self.speech_settings)
                else:
                    request.speech_settings = SpeechSettings(settings)
                request.message = line
                request.original_message = message
                request.message_id = message_id
                request.first_request_of_message = line == lines[0]
                request.last_request_of_message = line == lines[len(lines)-1]

                await self.process_queue.put(request)
        else:
            request = PendingSpeechRequest()

            if settings is None:
                request.speech_settings = SpeechSettings(self.speech_settings)
            else:
                request.speech_settings = SpeechSettings(settings)

            parts = self.__split_by_tags(message)
            requests: list[PendingSpeechRequest] = []
            tags: list[str] = []

            for part in parts:
                if part == '':
                    continue
                if part.startswith("<break"):
                    break_request = self.__create_ssml_request(request, part, tags)
                    if break_request is not None:
                        requests.append(break_request)
                elif part.startswith("</"):
                    tags.pop()
                elif part.startswith("<"):
                    tags.append(part)
                else:
                    lines = re.split(r'(?<=[.!?])\s+', part)
                    for line in lines:
                        requests.append(self.__create_ssml_request(request, line, tags))

            is_first: bool = True
            last_request = requests[len(requests)-1]
            for request in requests:
                request.first_request_of_message = is_first
                request.last_request_of_message = request == last_request
                request.original_message = message
                request.message_id = message_id

                is_first = False
                await self.process_queue.put(request)

    def stop_speaking(self):
        try:
            while not self.process_queue.empty():
                self.process_queue.get_nowait()
                self.process_queue.task_done()
            while not self.play_queue.empty():
                request = self.play_queue.get_nowait()
                f = Path(request.file_path)
                if f.exists():
                    os.remove(f)
                self.play_queue.task_done()
        except Exception as e:
            logging.info("Cleared speech queue")
            print("Cleared speech queue")

        self.stop_talking_event.set()

    def shutdown(self):
        self.shutdown_event.set()

    def set_volume(self, volume: float):
        self.volume = volume

    @staticmethod
    def __split_by_tags(html: str) -> list[str]:
        pattern = r"(</?[^>]+>)|([^<\n]+)"
        matches = re.findall(pattern, html)
        result = [match[0].strip() or match[1].strip() for match in matches if match[0] or match[1]]
        return result

    @staticmethod
    def __extract_attributes(tag) -> dict[str, str]:
        pattern = r'(\w+)=["\']([^"\']*?)["\']'
        attributes = dict(re.findall(pattern, tag))
        return attributes

    def __create_ssml_request(self, initial_request: PendingSpeechRequest, message: str, prosody_tags: list[str]) -> typing.Optional[PendingSpeechRequest]:
        request = PendingSpeechRequest()
        request.speech_settings = SpeechSettings(initial_request.speech_settings)

        if message.startswith("<break"):
            attr = self.__extract_attributes(message)
            if attr.__contains__("time"):
                if attr["time"].endswith("ms"):
                    request.silence_seconds = float(attr["time"].replace("ms", "")) / 1000
                elif attr["time"].endswith("s"):
                    request.silence_seconds = float(attr["time"].replace("s", ""))
                else:
                    request.silence_seconds = 1
            elif attr.__contains__("strength"):
                if attr["strength"] == "small" or attr["strength"] == "weak":
                    request.silence_seconds = .25
                elif attr["strength"] == "large":
                    request.silence_seconds = 2.5
                else:
                    request.silence_seconds = 1
            else:
                request.silence_seconds = 1
            request.silence_seconds = request.silence_seconds - .5
            if request.silence_seconds <= 0:
                return None
        else:
            for tag in prosody_tags:
                if tag.startswith("<prosody"):
                    attr = self.__extract_attributes(tag)
                    if attr.__contains__("volume"):
                        self.__apply_gain(request, attr["volume"])
                    if attr.__contains__("rate"):
                        self.__apply_rate(request, attr["rate"])
                    if attr.__contains__("pitch"):
                        self.__apply_pitch(request, attr["pitch"])
                elif tag.startswith("<voice"):
                    request.use_alt_voice = True

            request.message = message
        return request

    @staticmethod
    def __apply_gain(request: PendingSpeechRequest, volume: str):
        if volume.isnumeric():
            request.speech_settings.gain = float(volume) / 100.0 * 16 + -16
        elif volume == "silent":
            request.speech_settings.gain = -50
        elif volume == "x-soft":
            request.speech_settings.gain = -16
        elif volume == "soft":
            request.speech_settings.gain = -12
        elif volume == "medium":
            request.speech_settings.gain = -8
        elif volume == "loud":
            request.speech_settings.gain = -4
        else:
            request.speech_settings.gain = 0

    @staticmethod
    def __apply_rate(request: PendingSpeechRequest, rate: str):
        if rate.isnumeric():
            rate = float(rate)
            if rate < .5:
                request.speech_settings.speed = 0.5
            elif rate > 2:
                request.speech_settings.speed = 2
            else:
                request.speech_settings.speed = rate
        elif rate == "x-slow":
            request.speech_settings.speed = .5
        elif rate == "slow":
            request.speech_settings.speed = .8
        elif rate == "fast":
            request.speech_settings.speed = 1.55
        elif rate == "x-fast":
            request.speech_settings.speed *= 2

    @staticmethod
    def __apply_pitch(request: PendingSpeechRequest, pitch: str):
        if pitch.__contains__("%"):
            pitch = float(pitch.replace("%", "")) / 100.0
            if pitch < 0.5:
                request.speech_settings.pitch = 0.5
                request.speech_settings.speed *= 1.5
            elif pitch > 1.5:
                request.speech_settings.pitch = 1.5
                request.speech_settings.speed *= .5
            else:
                request.speech_settings.pitch = pitch
                request.speech_settings.speed *= 2 - pitch
        elif pitch == "x-low":
            request.speech_settings.pitch = 0.9
            request.speech_settings.speed *= 1.1
        elif pitch == "low":
            request.speech_settings.pitch = 0.95
            request.speech_settings.speed *= 1.05
        elif pitch == "high":
            request.speech_settings.pitch = 1.05
            request.speech_settings.speed *= 0.95
        elif pitch == "x-high":
            request.speech_settings.pitch = 1.1
            request.speech_settings.speed *= .9

    async def __send_response(self, request: PendingSpeechRequest, is_start: bool):
        if self.grpc_response_queue:
            response = speech_service_pb2.SpeechServiceResponse()
            response.speak_update.message = request.original_message
            if hasattr(request, "message") and request.message:
                response.speak_update.chunk = request.message
            elif hasattr(request, "silence_seconds") and request.silence_seconds:
                response.speak_update.chunk = "<break time='" + str(request.silence_seconds) + "s'/>"
            response.speak_update.is_start_of_message = is_start and request.first_request_of_message
            response.speak_update.is_start_of_chunk = is_start
            response.speak_update.is_end_of_message = not is_start and (request.last_request_of_message or self.stop_talking_event.is_set())
            response.speak_update.is_end_of_chunk = not is_start
            response.speak_update.has_another_request = not self.process_queue.empty() or not self.play_queue.empty()
            response.speak_update.message_id = request.message_id
            await self.grpc_response_queue.put(response)

    def __update_piper(self, request: PendingSpeechRequest) -> SpeechSettings:
        speech_settings = self.speech_settings
        if hasattr(request, "speech_settings"):
            speech_settings = request.speech_settings

        if hasattr(request, "use_alt_voice") and request.use_alt_voice:
            self.piper.set_speech_settings(speech_settings.alt_onnx_path, speech_settings.alt_config_path,
                                           speech_settings.alt_model_name)
        else:
            self.piper.set_speech_settings(speech_settings.onnx_path, speech_settings.config_path,
                                           speech_settings.model_name)

        return speech_settings

    def __load_audio_file(self, file: str):
        f = Path(file)
        try:
            return AudioSegment.from_file(f, format="wav")
        finally:
            if f.exists():
                os.remove(f)

    async def __handle_play(self, request: PendingSpeechRequest):
        if self.stop_talking_event.is_set():
            return
        self.is_speaking = True
        if hasattr(request, "message") and request.message:
            logging.debug("Playing " + request.message)
            await self.__send_response(request, True)

            speech_settings = self.speech_settings
            if hasattr(request, "speech_settings"):
                speech_settings = request.speech_settings
            sound = self.__load_audio_file(request.file_path)

            logging.debug("Loaded audio file " + request.file_path)

            frame_rate_modifier = 1

            if speech_settings.gain != 0:
                sound = sound.apply_gain(speech_settings.gain)
            if speech_settings.pitch != 1:
                frame_rate_modifier = speech_settings.pitch

            new_frame_rate = int(sound.frame_rate * frame_rate_modifier)
            sound = sound._spawn(sound.raw_data, overrides={"frame_rate": new_frame_rate})

            if self.volume != 1:
                sound = sound.apply_gain(ratio_to_db(self.volume))

            p = pyaudio.PyAudio()
            stream = p.open(format=p.get_format_from_width(sound.sample_width),
                            channels=sound.channels,
                            rate=sound.frame_rate,
                            output=True)

            logging.debug("Starting stream")

            try:
                await asyncio.to_thread(self.__write_sound_data, sound, stream)
            finally:
                await self.__send_response(request, False)
                stream.stop_stream()
                stream.close()
                p.terminate()
                self.is_speaking = False
                logging.info("Finished saying \"" + request.message + "\"")

            await asyncio.sleep(.5)

        elif hasattr(request, "silence_seconds") and request.silence_seconds:
            await asyncio.sleep(request.silence_seconds)
            self.is_speaking = False
            if request.last_request_of_message:
                await self.__send_response(request, False)

    async def __handle_request(self, request: PendingSpeechRequest):
        if hasattr(request, "message") and request.message:
            try:
                speech_settings = self.__update_piper(request)
                await asyncio.to_thread(self.piper.text_to_wav, request.message, request.file_path, speech_settings.speed)
                logging.info("Wrote to file " + request.file_path)
                await self.play_queue.put(request)
            except Exception as e:
                logging.error(e)
                logging.error(traceback.format_exc())
        elif hasattr(request, "silence_seconds") and request.silence_seconds:
            await self.play_queue.put(request)

    def __write_sound_data(self, sound: AudioSegment, stream: pyaudio.Stream):
        for chunk in make_chunks(sound, 500):
            if not self.stop_talking_event.is_set() and not self.shutdown_event.is_set():
                stream.write(chunk._data)


from enum import Enum

class GrammarElementType(Enum):
    Rule = 0
    String = 1
    OneOf = 2
    Optional = 3
    KeyValue = 4
    GrammarElementList = 5

class GrammarRuleElement:
    def __init(self, rule_name: str):
        self.rule = rule_name
        self.data = []

    rule: str = ""
    data = []

class GrammarElementLookupItem:

    def __init__(self, rule_name: str, phrase: str, element: GrammarRuleElement, full_match: bool):
        self.rule_name = rule_name
        self.grammar_element = element
        self.is_full_match = full_match
        self.phrase = phrase

    rule_name: str
    phrase: str
    grammar_element: GrammarRuleElement
    is_full_match: bool

class GrammarElementMatch:

    def __init__(self, rule_name: str, stated_text: str, matched_text: str, confidence: float, values=None):
        if values is None:
            values = {}
        self.rule = rule_name
        self.stated_text = stated_text
        self.matched_text = matched_text.strip()
        self.confidence = confidence
        self.values = values

    rule: str = ""
    stated_text: str = ""
    matched_text: str = ""
    confidence: float
    values: dict[str, str]


import asyncio
import datetime
import json
import logging
import sys
import time
import traceback
from asyncio import Server, Queue

from grpc import aio

from py_speech_service import speech_service_pb2_grpc, speech_service_pb2
from py_speech_service.speaker import Speaker, SpeechSettings
from py_speech_service.speech_recognition import SpeechRecognition
from py_speech_service.version import Version


class GrpcServer:

    speaker: Speaker
    speech_recognition: SpeechRecognition
    server: Server
    shutdown_event = asyncio.Event()
    response_queue = Queue()
    speech_initialized = False
    last_message = time.time()

    def __init__(self):
        self.speaker = Speaker()
        self.speech_recognition = SpeechRecognition()

    async def start(self):
        server = aio.server()
        self.server = server
        speech_service_pb2_grpc.add_SpeechServiceServicer_to_server(self, server)
        port = server.add_insecure_port("[::]:0")
        await server.start()
        logging.info("Listening to gRPC connections on port " + str(port))

        print(json.dumps({
            "version": Version.name(),
            "port": port
        }))

        asyncio.create_task(self.monitor())
        sys.stdout.flush()
        self.speaker.start()
        await self.shutdown_event.wait()
        self.speaker.shutdown()
        if self.speech_recognition:
            self.speech_recognition.shutdown()
        await server.stop(5)
        time.sleep(1)

    async def StartSpeechService(self, request_iterator, context):
        self.speaker.set_grpc_response_queue(self.response_queue)
        self.speech_recognition.set_grpc_response_queue(self.response_queue)
        asyncio.create_task(self.process_queue(context))
        self.last_message = time.time()

        async for request in request_iterator:
            try:
                self.last_message = time.time()

                logging.debug(str(request))

                if request.HasField("start_speech_recognition"):
                    logging.info("Received gRPC start_speech_recognition request")
                    print("Received gRPC start_speech_recognition request")

                    vosk_model = request.start_speech_recognition.vosk_model if hasattr(request.start_speech_recognition, "vosk_model") else None
                    grammar_file = request.start_speech_recognition.grammar_file if hasattr(request.start_speech_recognition, "grammar_file") else None
                    required_confidence = request.start_speech_recognition.required_confidence if hasattr(request.start_speech_recognition, "required_confidence") else 80

                    successful = self.speech_recognition.set_speech_recognition_details(grammar_file, vosk_model, required_confidence)
                    if successful:
                        asyncio.create_task(self.speech_recognition.start_speech_recognition(context))
                    else:
                        response = speech_service_pb2.SpeechServiceResponse()
                        response.speech_recognition_started.successful = False
                        await self.response_queue.put(response)

                elif request.HasField("set_speech_settings"):
                    logging.info(
                        "Received gRPC set_speech_settings request: " + str(request.set_speech_settings))
                    print("Received gRPC set_speech_settings request")

                    try:
                        self.speaker.init_speech_settings(SpeechSettings(request.set_speech_settings.speech_settings))
                        self.speech_initialized = True
                    except Exception as e:
                        response = speech_service_pb2.SpeechServiceResponse()
                        response.speech_settings_set.successful = True
                        await self.response_queue.put(response)
                        logging.error("Error initializing speech settings: " + repr(e))
                        logging.error(traceback.format_exc())
                        response = speech_service_pb2.SpeechServiceResponse()
                        response.error.error_message = "Error initializing speech settings"
                        response.error.exception = repr(e)
                        await self.response_queue.put(response)

                elif request.HasField("speak"):
                    logging.info(
                        "Received gRPC speak request: \"" + ' '.join(request.speak.message.splitlines()) + "\"")
                    print("Received gRPC speak request")

                    if self.speech_initialized:
                        speech_settings = SpeechSettings(request.speak.speech_settings) if request.speak.HasField("speech_settings") else None
                        await self.speaker.speak(request.speak.message, speech_settings, request.speak.message_id)
                    else:
                        response = speech_service_pb2.SpeechServiceResponse()
                        response.error.error_message = "Speech settings have not been initialized. Call set_speech_settings first."
                        await self.response_queue.put(response)

                elif request.HasField("stop_speaking"):
                    logging.info(
                        "Received gRPC stop_speaking request")
                    print("Received gRPC stop_speaking request")

                    self.speaker.stop_speaking()
                elif request.HasField("shutdown"):
                    logging.info(
                        "Received gRPC shutdown request")
                    print("Received gRPC shutdown request")
                    break
                elif request.HasField("ping"):
                    print("Received gRPC ping")
                    logging.info(
                        "Received gRPC ping")
                    response = speech_service_pb2.SpeechServiceResponse()
                    response.ping.time = str(datetime.datetime.now())
                    await self.response_queue.put(response)
                elif request.HasField("stop_speech_recognition"):
                    print("Received stop speech recognition request")
                    self.speech_recognition.stop_speech_recognition()
                elif request.HasField("set_volume"):
                    logging.info("Received set volume request")
                    self.speaker.set_volume(request.set_volume.volume)
                    response = speech_service_pb2.SpeechServiceResponse()
                    response.set_volume.successful = True
                    await self.response_queue.put(response)

            except Exception as e:
                logging.error("Exception with speech service: " + str(e))
                logging.error(repr(e))
                logging.error(traceback.format_exc())
                response = speech_service_pb2.SpeechServiceResponse()
                response.error.error_message = "Exception with speech service: " + str(e)
                response.error.exception = repr(e)
                await self.response_queue.put(response)

        self.shutdown_event.set()

    async def process_queue(self, context):
        while not self.shutdown_event.is_set():
            response = await self.response_queue.get()
            await context.write(response)

    async def monitor(self):
        logging.info("Starting monitor")
        while time.time() - self.last_message < 300 and not self.shutdown_event.is_set():
            await asyncio.sleep(60)

        if not self.shutdown_event.is_set():
            logging.info("No message received in 5 minutes. Exiting.")
            print("No message received in 5 minutes. Exiting.")
            self.shutdown_event.set()





import asyncio
import logging
import os
import sys
import time
import traceback
from logging.handlers import RotatingFileHandler
from pathlib import Path

from platformdirs import user_data_dir

from py_speech_service.grpc_server import GrpcServer
from py_speech_service.speaker import Speaker, SpeechSettings
from py_speech_service.speech_recognition import SpeechRecognition
from py_speech_service.version import Version


def cli():
    try:

        log_path = get_arg_value("-l")
        if not log_path:
            log_path = os.path.join(user_data_dir("py_speech_service"), "log.log")

        Path(os.path.dirname(log_path)).mkdir(parents=True, exist_ok=True)

        logging.basicConfig(
            handlers=[RotatingFileHandler(log_path, maxBytes=500000, backupCount=3)],
            level=logging.DEBUG if get_arg_flag("-d") else logging.INFO,
            format="[%(asctime)s] %(levelname)s [%(name)s.%(funcName)s:%(lineno)d] %(message)s",
            datefmt='%Y-%m-%d %H:%M:%S')

        logging.info("Starting py-speech-service v" + Version.name())

        arg_array = sys.argv

        first_arg = arg_array[0]
        second_arg = arg_array[1] if len(arg_array) > 1 else 0

        if first_arg == "version" or second_arg == "version" or first_arg == "-v" or second_arg == "-v" or first_arg == "--version" or second_arg == "--version":
            logging.info("Returning version")
            print("py-speech-service v" + Version.name())
        elif first_arg == "speak" or second_arg == "speak":
            logging.info("Starting speak mode")
            speech = arg_array.pop()
            speaker = Speaker()
            speaker.init_speech_settings(SpeechSettings())
            asyncio.run(speaker.speak_basic_line(speech))
        elif first_arg == "recognition" or second_arg == "recognition":
            logging.info("Starting speech recognition mode")

            grammar = ""
            model = ""
            for arg in arg_array:
                if arg.startswith("-g="):
                    grammar = get_single_arg(arg, "-g")
                elif arg.startswith("-m"):
                    model = get_single_arg(arg, "-m")

            logging.info("Grammar file: " + grammar)
            logging.info("model: " + model)

            speech_recognition = SpeechRecognition()
            speech_recognition.set_speech_recognition_details(grammar, model)
            asyncio.run(speech_recognition.start_speech_recognition(None))

        elif first_arg == "service" or second_arg == "service":
            logging.info("Starting gRPC server mode")
            server = GrpcServer()
            asyncio.run(server.start())
        else:
            logging.info("Printing documentation")
            print("py-speech-service v" + Version.name())
            print("Usage: py-speech-service (speak/recognition/service)")
            print("  py-speech-service speak \"text to speech\"")
            print("  py-speech-service recognition -g \"path to grammar file\" -m \"path to VOSK model folder\"")
            print("  py-speech-service service -g \"path to grammar file\" -m \"path to VOSK model folder\" -p \"preferred port\"")

    except Exception as e:
        logging.error(e)
        logging.error(traceback.format_exc())

    time.sleep(5)

def get_arg_value(arg: str):
    for current_arg in sys.argv:
        if current_arg.startswith(arg + "="):
            return get_single_arg(current_arg, arg)
    return None

def get_arg_flag(arg: str):
    for current_arg in sys.argv:
        if current_arg == arg:
            return True
    return False

def get_single_arg(arg: str, command: str):
    if arg.startswith(command + "=\""):
        return arg[len(command)+2:-1]
    else:
        return arg[len(command)+1:]

if __name__ == "__main__":
    cli()



import json
import logging
import random
import re
import typing

import num2words
from rapidfuzz import process

from py_speech_service.grammar_element import GrammarRuleElement, GrammarElementType, GrammarElementLookupItem, \
    GrammarElementMatch


class GrammarParser:

    phrase_map: dict[str, list[GrammarElementLookupItem]] = {}
    leading_phrases: list[str] = []
    pattern = re.compile('[^\\w ]+')
    max_phrase_word_count: int = 0
    all_words: list[str] = []
    phrase_replacement_map: dict[str, list[str]]
    phrase_replacement_regex: str
    replacement_map: dict[str, str]
    replacement_regex: str
    has_replacements: bool = False
    prefix: list[str]

    def set_grammar_file(self, file_path: str):
        with open(file_path, 'r') as fp:
            lines = fp.read()

        self.all_words = []
        self.replacement_map = {}
        self.replacement_regex = ""
        self.phrase_replacement_map = {}
        self.phrase_replacement_regex = ""

        json_data = json.loads(lines)

        logging.info("Loaded grammar json data file " + file_path)

        if "Replacements" in json_data and json_data["Replacements"] and len(json_data["Replacements"]) > 0:
            for find_text, replace_with in json_data["Replacements"].items():
                find_text_str = str(find_text).lower()
                replace_with_str = str(replace_with).lower()
                self.replacement_map[find_text_str] = replace_with_str
                self.all_words.append(find_text_str)

                if not self.phrase_replacement_map.__contains__(replace_with_str):
                    self.phrase_replacement_map[replace_with_str] = []
                self.phrase_replacement_map[replace_with_str].append(find_text_str)

            logging.info("Loaded " + str(len(json_data["Replacements"])) + " replacements")

            sorted_replacement_keys = sorted(self.replacement_map.keys(), key=len, reverse=True)
            self.replacement_regex = r'(' + '|'.join(map(re.escape, sorted_replacement_keys)) + r')'

            sorted_replacement_keys = sorted(self.phrase_replacement_map.keys(), key=len, reverse=True)
            self.phrase_replacement_regex = r'\b(' + '|'.join(map(re.escape, sorted_replacement_keys)) + r')\b'

        for rule in json_data["Rules"]:
            rule_name: str = rule["Key"]
            self.__parse_rule_element(rule_name, rule)

        logging.info("Loaded " + str(len(json_data["Rules"])) + " rules")

        if "Prefix" in json_data:
            self.prefix = json_data["Prefix"].lower().split()
        else:
            self.prefix = []

        self.all_words = list(set(self.all_words))
        updated_words = []
        for word in self.all_words:
            updated_words += self.pattern.sub('', word).lower().split()
        self.all_words = list(set(updated_words))

    def find_match(self, stated_text: str, min_threshold: float = 80, min_prefix_threshold: float = 60):
        search_text = self.pattern.sub('', stated_text).lower().strip()
        search_words = search_text.split()
        search_word_count = len(search_words)
        if search_word_count > 30:
            return None

        if len(self.prefix) > 0:
            prefix_search = " ".join(search_words[:len(self.prefix)])
            prefix_result = self.__find_closest_sentence([" ".join(self.prefix)], prefix_search)
            if prefix_result is None:
                return None
            if prefix_result[1] < min_prefix_threshold:
                return None
            index = 0
            for word in self.prefix:
                search_words[index] = word
                index = index + 1
            search_text = " ".join(search_words)
            logging.info("Matched prefix " + " ".join(self.prefix))

        possibilities: [(str, float)] = []
        for i in range(search_word_count, 1, -1):
            search_phrase = " ".join(search_words[0:i])
            search_result = self.__find_closest_sentence(self.leading_phrases, search_phrase)
            if search_result is not None and search_result[1] > min_threshold:
                possibilities.append(search_result)

        matches: dict[str, GrammarElementMatch] = {}
        searched_phrases: [str] = []
        for possibility in possibilities:
            search_phrase = possibility[0]
            if searched_phrases.__contains__(search_phrase):
                continue
            searched_phrases.append(search_phrase)
            initial_confidence = possibility[1]
            possible_elements = self.phrase_map.get(search_phrase)
            searched_elements = []
            for possible_element in possible_elements:
                if possible_element.is_full_match:
                    if initial_confidence > min_threshold:
                        match = GrammarElementMatch(possible_element.rule_name, stated_text, search_phrase, initial_confidence)
                        matches[match.matched_text] = match
                else:
                    best_item = self.__find_best_element(search_text, search_phrase, possible_element)
                    if best_item is not None and best_item[1] > min_threshold:
                        match = GrammarElementMatch(possible_element.rule_name, stated_text, best_item[0],
                                                    best_item[1], best_item[2])
                        matches[match.matched_text] = match

        if len(matches) == 0:
            return None
        else:
            closest_sentence = self.__find_closest_sentence(matches.keys(), search_text)
            if closest_sentence is None:
                return None
            selected_match = matches[closest_sentence[0]]
            selected_match.confidence = (closest_sentence[1] + selected_match.confidence) / 2

            if selected_match.confidence > 98:
                selected_match.confidence = selected_match.confidence - random.uniform(0.5, 2.5)

            if self.replacement_regex:
                selected_match.matched_text = re.sub(self.replacement_regex, lambda regex_match: self.replacement_map[regex_match.group(0)],
                                     selected_match.matched_text)

            return selected_match

    def __parse_rule_element(self, rule_name: str, rule_element):

        element = GrammarRuleElement()
        element.rule = rule_name
        element.data = []
        element_phrases: list[str] = [""]

        is_exact: bool = True
        words = []

        for sub_element_json in rule_element['Data']:
            element.data.append(sub_element_json)

            element_type: GrammarElementType = GrammarElementType(sub_element_json['Type'])

            if element_type == GrammarElementType.String:
                for i in range(0, len(element_phrases)):
                    text = self.pattern.sub('', sub_element_json['Data'].strip() + " ").lower()
                    words += [ text]
                    if is_exact:
                        element_phrases[i] += text
            elif element_type == GrammarElementType.OneOf:
                one_of_phrases: [str] = sub_element_json['Data']
                words += one_of_phrases
                if is_exact:
                    element_phrases = self.__permutate_elements(element_phrases, one_of_phrases)
            elif element_type == GrammarElementType.Optional:
                optional_phrases: [str] = sub_element_json['Data']
                words += optional_phrases
                if is_exact:
                    optional_phrases.append("")
                    element_phrases = self.__permutate_elements(element_phrases, optional_phrases)
            elif element_type == GrammarElementType.KeyValue:
                items = []
                for key_value_json in sub_element_json['Data']:
                    key = key_value_json['Key']
                    if key.isnumeric():
                        key = num2words.num2words(key).replace("-", " ")
                    key = str(key).lower()

                    if self.phrase_replacement_regex:
                        match = re.search(self.phrase_replacement_regex, key)
                        if match:
                            replacements = self.phrase_replacement_map[match.group()]
                            for replacement in replacements:
                                new_key = str(key).replace(match.group(), replacement)
                                items.append({ "Key": new_key, "Value" : key_value_json['Value']})
                                words.append(new_key)
                        else:
                            items.append(key_value_json)
                            words.append(key)
                    else:
                        items.append(key_value_json)
                        words.append(key)
                sub_element_json['Data'] = items
                is_exact = False
            elif element_type == GrammarElementType.GrammarElementList:
                for grammar_list_item in sub_element_json['Data']:
                    self.__parse_rule_element(rule_name, grammar_list_item)
                return

        self.all_words += words
        for phrase in element_phrases:
            match_details = GrammarElementLookupItem(rule_name, phrase, element, is_exact)
            word_count = len(phrase.split())
            if phrase.strip() == "" or word_count < 2:
                continue
            if word_count > self.max_phrase_word_count:
                self.max_phrase_word_count = word_count
            if self.phrase_map.__contains__(phrase):
                self.phrase_map[phrase].append(match_details)
            else:
                self.phrase_map[phrase] = [ match_details ]
                self.leading_phrases.append(phrase)

    def __permutate_elements(self, initial_items: [str], additional_items: [str]):
        initial_count = len(initial_items)
        to_return = self.__duplicate_elements(initial_items, len(additional_items))
        index = 0
        for i in range(0, initial_count):
            for text in additional_items:
                text = self.pattern.sub('', text.strip() + " ").lower()
                if text == " ":
                    index = index + 1
                    continue
                to_return[index] += text
                index = index + 1
        return to_return

    def __find_best_element(self, search_text: str, current_phrase: str, element: GrammarElementLookupItem):
        hit_key_value = False
        search_phrase = current_phrase
        selected_values: dict[str, str] = {}
        confidence = 0
        search_words = search_text.split()

        for sub_element_json in element.grammar_element.data:
            element_type: GrammarElementType = GrammarElementType(sub_element_json['Type'])
            if not hit_key_value:
                if element_type == GrammarElementType.KeyValue:
                    hit_key_value = True
                else:
                    continue

            if element_type == GrammarElementType.String:
                text = self.pattern.sub('', sub_element_json['Data'].strip() + " ").lower()
                search_phrase += text
            elif element_type == GrammarElementType.OneOf:
                items: [str] = sub_element_json['Data']
                match = self.__find_best_element_sub_phrase(search_phrase, search_words, items)
                if match is None:
                    return None
                else:
                    search_phrase = match[0]
                    confidence = match[1]
            elif element_type == GrammarElementType.Optional:
                items: [str] = sub_element_json['Data']
                match = self.__find_best_element_sub_phrase(search_phrase, search_words, items)
                if match is not None and match[1] > confidence:
                    search_phrase = match[0]
                    confidence = match[1]
            elif element_type == GrammarElementType.KeyValue:
                keys: [str] = []
                key_values: dict[str, str] = {}
                for key_value_json in sub_element_json['Data']:
                    key = key_value_json['Key']
                    if key.isnumeric():
                        key = num2words.num2words(key).replace("-", " ")
                    keys.append(key)
                    key_values[key] = key_value_json['Value']
                match = self.__find_best_element_sub_phrase(search_phrase, search_words, keys)
                if match is None:
                    return None
                else:
                    search_phrase = match[0]
                    confidence = match[1]
                    selected_values[str(sub_element_json['Key'])] = str(key_values[match[2]])

        return search_phrase, confidence, selected_values

    def __find_best_element_sub_phrase(self, search_phrase: str, search_words: [str], items: [str]):
        if not search_phrase.endswith(" "):
            search_phrase = search_phrase + " "
        possible_phrases: [str] = []
        phrase_addition_map: dict[str, str] = {}
        max_word_count = 0
        min_word_count = 10000
        for item in items:
            text = self.pattern.sub('', item.strip() + " ").lower()
            new_phrase = search_phrase + text
            possible_phrases.append(new_phrase)
            phrase_addition_map[new_phrase] = item
            new_phrase_word_count = len(new_phrase.split())
            if new_phrase_word_count < min_word_count:
                min_word_count = new_phrase_word_count
            if new_phrase_word_count > max_word_count:
                max_word_count = new_phrase_word_count

        best_result = None
        previous_search_phrase = ""
        for num_words in range(min_word_count, max_word_count+2):
            new_search_phrase = " ".join(search_words[0:num_words])
            if new_search_phrase == previous_search_phrase:
                break
            previous_search_phrase = new_search_phrase
            result = self.__find_closest_sentence(possible_phrases, new_search_phrase)
            if result is not None:
                if best_result is None:
                    best_result = result
                elif result[1] >= best_result[1] - 2:
                    best_result = result

        if best_result is None:
            return None

        return best_result[0], best_result[1], phrase_addition_map[best_result[0]]

    def __find_closest_sentence(self, sentences: [str], query: str) -> (str, float):
        filtered_sentences = self.__filter_by_length(sentences, query)
        squashed_sentences = []
        squashed_map = {}
        for sentence in filtered_sentences:
            squashed_sentence = sentence.replace(" ", "")
            squashed_sentences.append(squashed_sentence)
            squashed_map[squashed_sentence] = sentence
        response = process.extractOne(query.replace(" ", ""), squashed_sentences)
        if response is None:
            return None
        return squashed_map[response[0]], response[1]

    @staticmethod
    def __filter_by_length(items: [], reference_string: str, tolerance: int = 4):
        reference_length = len(reference_string)
        return [s for s in items if abs(len(s) - reference_length) <= tolerance]

    @staticmethod
    def __duplicate_elements(items: [], times: int):
        if times < 1:
            return []  # Return an empty list if times is less than 1
        return [element for element in items for _ in range(times)]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: speech_service.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x14speech_service.proto\"\xa8\x03\n\x14SpeechServiceRequest\x12\x42\n\x18start_speech_recognition\x18\x01 \x01(\x0b\x32\x1e.StartSpeechRecognitionRequestH\x00\x12\x38\n\x13set_speech_settings\x18\x02 \x01(\x0b\x32\x19.SetSpeechSettingsRequestH\x00\x12\x1e\n\x05speak\x18\x03 \x01(\x0b\x32\r.SpeakRequestH\x00\x12-\n\rstop_speaking\x18\x04 \x01(\x0b\x32\x14.StopSpeakingRequestH\x00\x12$\n\x08shutdown\x18\x05 \x01(\x0b\x32\x10.ShutdownRequestH\x00\x12\x1c\n\x04ping\x18\x06 \x01(\x0b\x32\x0c.PingRequestH\x00\x12@\n\x17stop_speech_recognition\x18\x07 \x01(\x0b\x32\x1d.StopSpeechRecognitionRequestH\x00\x12-\n\nset_volume\x18\x08 \x01(\x0b\x32\x17.SetSpeechVolumeRequestH\x00\x42\x0e\n\x0cmessage_type\"\x85\x03\n\x15SpeechServiceResponse\x12,\n\x0cspeak_update\x18\x01 \x01(\x0b\x32\x14.SpeakUpdateResponseH\x00\x12\x37\n\x11speech_recognized\x18\x02 \x01(\x0b\x32\x1a.SpeechRecognitionResponseH\x00\x12$\n\x05\x65rror\x18\x03 \x01(\x0b\x32\x13.SpeechServiceErrorH\x00\x12\x1d\n\x04ping\x18\x04 \x01(\x0b\x32\r.PingResponseH\x00\x12\x45\n\x1aspeech_recognition_started\x18\x05 \x01(\x0b\x32\x1f.StartSpeechRecognitionResponseH\x00\x12\x39\n\x13speech_settings_set\x18\x06 \x01(\x0b\x32\x1a.SetSpeechSettingsResponseH\x00\x12.\n\nset_volume\x18\x07 \x01(\x0b\x32\x18.SetSpeechVolumeResponseH\x00\x42\x0e\n\x0cmessage_type\">\n\x12SpeechServiceError\x12\x15\n\rerror_message\x18\x01 \x01(\t\x12\x11\n\texception\x18\x02 \x01(\t\"f\n\x1dStartSpeechRecognitionRequest\x12\x12\n\nvosk_model\x18\x01 \x01(\t\x12\x14\n\x0cgrammar_file\x18\x02 \x01(\t\x12\x1b\n\x13required_confidence\x18\x03 \x01(\x01\"\x1e\n\x1cStopSpeechRecognitionRequest\"D\n\x18SetSpeechSettingsRequest\x12(\n\x0fspeech_settings\x18\x01 \x01(\x0b\x32\x0f.SpeechSettings\"\xc0\x01\n\x0eSpeechSettings\x12\x12\n\nmodel_name\x18\x01 \x01(\t\x12\x11\n\tonnx_path\x18\x02 \x01(\t\x12\x13\n\x0b\x63onfig_path\x18\x03 \x01(\t\x12\x16\n\x0e\x61lt_model_name\x18\x04 \x01(\t\x12\x15\n\ralt_onnx_path\x18\x05 \x01(\t\x12\x17\n\x0f\x61lt_config_path\x18\x06 \x01(\t\x12\r\n\x05speed\x18\x07 \x01(\x01\x12\x0c\n\x04gain\x18\x08 \x01(\x01\x12\r\n\x05pitch\x18\t \x01(\x01\"v\n\x0cSpeakRequest\x12\x0f\n\x07message\x18\x01 \x01(\t\x12-\n\x0fspeech_settings\x18\x02 \x01(\x0b\x32\x0f.SpeechSettingsH\x00\x88\x01\x01\x12\x12\n\nmessage_id\x18\x03 \x01(\x04\x42\x12\n\x10_speech_settings\"\x15\n\x13StopSpeakingRequest\"\xd2\x01\n\x13SpeakUpdateResponse\x12\x0f\n\x07message\x18\x01 \x01(\t\x12\r\n\x05\x63hunk\x18\x02 \x01(\t\x12\x1b\n\x13is_start_of_message\x18\x03 \x01(\x08\x12\x19\n\x11is_start_of_chunk\x18\x04 \x01(\x08\x12\x19\n\x11is_end_of_message\x18\x05 \x01(\x08\x12\x17\n\x0fis_end_of_chunk\x18\x06 \x01(\x08\x12\x1b\n\x13has_another_request\x18\x07 \x01(\x08\x12\x12\n\nmessage_id\x18\x08 \x01(\x04\"\xe5\x01\n\x19SpeechRecognitionResponse\x12\x12\n\nheard_text\x18\x01 \x01(\t\x12\x17\n\x0frecognized_text\x18\x02 \x01(\t\x12\x17\n\x0frecognized_rule\x18\x03 \x01(\t\x12\x12\n\nconfidence\x18\x04 \x01(\x01\x12<\n\tsemantics\x18\x05 \x03(\x0b\x32).SpeechRecognitionResponse.SemanticsEntry\x1a\x30\n\x0eSemanticsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\x11\n\x0fShutdownRequest\"\x1b\n\x0bPingRequest\x12\x0c\n\x04time\x18\x01 \x01(\t\"\x1c\n\x0cPingResponse\x12\x0c\n\x04time\x18\x01 \x01(\t\"4\n\x1eStartSpeechRecognitionResponse\x12\x12\n\nsuccessful\x18\x01 \x01(\x08\"/\n\x19SetSpeechSettingsResponse\x12\x12\n\nsuccessful\x18\x01 \x01(\x08\"(\n\x16SetSpeechVolumeRequest\x12\x0e\n\x06volume\x18\x01 \x01(\x01\"-\n\x17SetSpeechVolumeResponse\x12\x12\n\nsuccessful\x18\x01 \x01(\x08\x32X\n\rSpeechService\x12G\n\x12StartSpeechService\x12\x15.SpeechServiceRequest\x1a\x16.SpeechServiceResponse(\x01\x30\x01\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'speech_service_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _SPEECHRECOGNITIONRESPONSE_SEMANTICSENTRY._options = None
  _SPEECHRECOGNITIONRESPONSE_SEMANTICSENTRY._serialized_options = b'8\001'
  _SPEECHSERVICEREQUEST._serialized_start=25
  _SPEECHSERVICEREQUEST._serialized_end=449
  _SPEECHSERVICERESPONSE._serialized_start=452
  _SPEECHSERVICERESPONSE._serialized_end=841
  _SPEECHSERVICEERROR._serialized_start=843
  _SPEECHSERVICEERROR._serialized_end=905
  _STARTSPEECHRECOGNITIONREQUEST._serialized_start=907
  _STARTSPEECHRECOGNITIONREQUEST._serialized_end=1009
  _STOPSPEECHRECOGNITIONREQUEST._serialized_start=1011
  _STOPSPEECHRECOGNITIONREQUEST._serialized_end=1041
  _SETSPEECHSETTINGSREQUEST._serialized_start=1043
  _SETSPEECHSETTINGSREQUEST._serialized_end=1111
  _SPEECHSETTINGS._serialized_start=1114
  _SPEECHSETTINGS._serialized_end=1306
  _SPEAKREQUEST._serialized_start=1308
  _SPEAKREQUEST._serialized_end=1426
  _STOPSPEAKINGREQUEST._serialized_start=1428
  _STOPSPEAKINGREQUEST._serialized_end=1449
  _SPEAKUPDATERESPONSE._serialized_start=1452
  _SPEAKUPDATERESPONSE._serialized_end=1662
  _SPEECHRECOGNITIONRESPONSE._serialized_start=1665
  _SPEECHRECOGNITIONRESPONSE._serialized_end=1894
  _SPEECHRECOGNITIONRESPONSE_SEMANTICSENTRY._serialized_start=1846
  _SPEECHRECOGNITIONRESPONSE_SEMANTICSENTRY._serialized_end=1894
  _SHUTDOWNREQUEST._serialized_start=1896
  _SHUTDOWNREQUEST._serialized_end=1913
  _PINGREQUEST._serialized_start=1915
  _PINGREQUEST._serialized_end=1942
  _PINGRESPONSE._serialized_start=1944
  _PINGRESPONSE._serialized_end=1972
  _STARTSPEECHRECOGNITIONRESPONSE._serialized_start=1974
  _STARTSPEECHRECOGNITIONRESPONSE._serialized_end=2026
  _SETSPEECHSETTINGSRESPONSE._serialized_start=2028
  _SETSPEECHSETTINGSRESPONSE._serialized_end=2075
  _SETSPEECHVOLUMEREQUEST._serialized_start=2077
  _SETSPEECHVOLUMEREQUEST._serialized_end=2117
  _SETSPEECHVOLUMERESPONSE._serialized_start=2119
  _SETSPEECHVOLUMERESPONSE._serialized_end=2164
  _SPEECHSERVICE._serialized_start=2166
  _SPEECHSERVICE._serialized_end=2254
# @@protoc_insertion_point(module_scope)



import logging
import os
import ssl
import sys
import tempfile
import urllib
from pathlib import Path
from urllib.request import urlretrieve

import certifi
import yapper.constants as c
import requests
import zipfile
import tarfile
import platform

from yapper import PiperVoiceUS, PiperQuality, PiperVoiceUK

if os.name == "nt":
    PLATFORM = c.PLATFORM_WINDOWS
elif os.name == "posix":
    home = Path.home()
    if os.uname().sysname == "Darwin":
        PLATFORM = c.PLATFORM_MAC
    else:
        PLATFORM = c.PLATFORM_LINUX
else:
    print("your system is not supported")
    sys.exit()

def download_file(url: str) -> str:
    # Create a temporary file
    temp_file = tempfile.NamedTemporaryFile(delete=False)
    try:
        print(f"Downloading {url} ...")
        # Stream the download to avoid loading it into memory all at once
        response = requests.get(url, stream=True)
        response.raise_for_status()

        # Write the content to the temp file
        with open(temp_file.name, 'wb') as f:
            for chunk in response.iter_content(chunk_size=8192):
                f.write(chunk)

        print(f"Downloaded to {temp_file.name}")
        return temp_file.name
    except Exception as e:
        print(f"Error downloading file: {e}")
        raise


def extract_zip(source_file, dest_folder):
    print(f"Extracting {source_file} to {dest_folder} ...")
    # Create the destination folder if it doesn't exist
    os.makedirs(dest_folder, exist_ok=True)

    # Extract the ZIP file
    with zipfile.ZipFile(source_file, 'r') as zip_ref:
        zip_ref.extractall(dest_folder)

    print(f"Extraction complete. Files are in {dest_folder}")


def extract_tar_gz(source_file, dest_folder):
    print(f"Extracting TAR.GZ: {source_file} to {dest_folder} ...")
    # Create the destination folder if it doesn't exist
    os.makedirs(dest_folder, exist_ok=True)

    # Extract the TAR.GZ file
    with tarfile.open(source_file, 'r:gz') as tar_ref:
        tar_ref.extractall(dest_folder)

    print(f"TAR.GZ extraction complete. Files are in {dest_folder}")


def download_and_extract(url: str, destination_folder: str):
    downloaded_file: str = ""
    try:
        downloaded_file = download_file(url)
        if url.lower().endswith(".zip"):
            extract_zip(downloaded_file, destination_folder)
        elif url.lower().endswith(".tar.gz"):
            extract_tar_gz(downloaded_file, destination_folder)
        else:
            raise ValueError("Unsupported file type. Only .zip and .tar.gz are supported.")
    finally:
        if downloaded_file and os.path.exists(downloaded_file):
            os.remove(downloaded_file)
            print(f"Temporary file {downloaded_file} deleted.")

def get_json_data(url):
    try:
        print(f"Fetching JSON data from {url} ...")
        response = requests.get(url)
        response.raise_for_status()
        json_data = response.json()
        print("Data fetched successfully.")
        return json_data
    except requests.exceptions.RequestException as e:
        print(f"Error fetching data: {e}")
        raise

def download(url: str, file: str):
    logging.info(f"Downloading {url} to {file}")
    context = ssl.create_default_context(cafile=certifi.where())
    # urllib.request.urlretrieve(url, file, context=context)
    with urllib.request.urlopen(url, context=context) as response, open(file, 'wb') as out_file:
        out_file.write(response.read())
    logging.info(f"Download of {url} complete")

def download_piper(piper_dir: Path):

    """Installs piper into the app's home directory."""
    if (piper_dir / "piper").exists():
        return
    zip_path = piper_dir / "piper.zip"
    prefix = "https://github.com/rhasspy/piper/releases/download/2023.11.14-2"
    if PLATFORM == c.PLATFORM_LINUX:
        if platform.machine() in ("aarch64", "arm64"):
            nix_link = f"{prefix}/piper_linux_aarch64.tar.gz"
        elif platform.machine() in ("armv7l", "armv7"):
            nix_link = f"{prefix}/piper_linux_armv7l.tar.gz"
        else:
            nix_link = f"{prefix}/piper_linux_x86_64.tar.gz"
        download(nix_link, str(zip_path))
    elif PLATFORM == c.PLATFORM_WINDOWS:
        download(f"{prefix}/piper_windows_amd64.zip", str(zip_path))
    else:
        download(f"{prefix}/piper_macos_x64.tar.gz", str(zip_path))

    if PLATFORM == c.PLATFORM_WINDOWS:
        with zipfile.ZipFile(zip_path, "r") as z_f:
            z_f.extractall(piper_dir)
    else:
        with tarfile.open(zip_path, "r") as z_f:
            z_f.extractall(piper_dir)
    os.remove(zip_path)


def download_piper_model(
    voice: PiperVoiceUS | PiperVoiceUK,
    quality: PiperQuality,
    piper_dir: Path
) -> tuple[str, str]:
    voices_dir = piper_dir / "piper_voices"
    voices_dir.mkdir(exist_ok=True)
    lang_code = "en_US" if isinstance(voice, PiperVoiceUS) else "en_GB"
    voice, quality = voice.value, quality.value

    onnx_file = voices_dir / f"{lang_code}-{voice}-{quality}.onnx"
    conf_file = voices_dir / f"{lang_code}-{voice}-{quality}.onnx.json"

    prefix = "https://huggingface.co/rhasspy/piper-voices/resolve/main/en/"
    prefix += lang_code
    help_url = "https://huggingface.co/rhasspy/piper-voices/tree/main/en/"
    help_url += lang_code
    if not onnx_file.exists():
        try:
            onnx_url = (
                f"{prefix}/{voice}/{quality}/{onnx_file.name}?download=true"
            )
            download(onnx_url, str(onnx_file))
        except (KeyboardInterrupt, Exception) as e:
            onnx_file.unlink(missing_ok=True)
            if getattr(e, "status", None) == 404:
                raise Exception(
                    f"{voice}({quality}) is not available, please refer to"
                    f" {help_url} to check all available models"
                )
            raise e
    if not conf_file.exists():
        conf_url = f"{prefix}/{voice}/{quality}/{conf_file.name}?download=true"
        try:
            download(conf_url, str(conf_file))
        except (KeyboardInterrupt, Exception) as e:
            conf_file.unlink(missing_ok=True)
            raise e

    return str(onnx_file), str(conf_file)



import asyncio
import json
import logging
import os
import queue
import traceback
from typing import Optional

import sounddevice as sd
from platformdirs import user_data_dir
from vosk import Model, KaldiRecognizer, SetLogLevel

from py_speech_service import speech_service_pb2
from py_speech_service.downloader import get_json_data, download_and_extract
from py_speech_service.grammar_parser import GrammarParser


class SpeechRecognition:
    model: Model
    grammar_parser: GrammarParser
    required_confidence: float = 80
    stop_speech_recognition_event: Optional[asyncio.Event] = None
    continue_speech_recognition: bool = True
    grpc_response_queue: Optional[asyncio.Queue] = None
    shutdown_event = asyncio.Event()
    recognition_queue = asyncio.Queue()
    vosk_model_folder = os.path.join(user_data_dir("py_speech_service"), "vosk")

    def __init__(self):
        SetLogLevel(-1)
        self.stop_speech_recognition_event = None

    def download_vosk_model(self, model_name: Optional[str]) -> str:
        model = model_name
        if not model:
            model = "vosk-model-small-en-us-0.15"
        if os.path.exists(self.get_vosk_model_path(model)):
            return self.get_vosk_model_path(model)
        url = self.get_vosk_download_url_by_name(model)
        download_and_extract(url, self.vosk_model_folder)
        return self.get_vosk_model_path(model)

    def get_vosk_model_path(self, model_name: Optional[str]) -> str:
        return os.path.join(self.vosk_model_folder, model_name)

    def get_vosk_download_url_by_name(self, vosk_model_name) -> str:
        try:
            json_data = get_json_data("https://alphacephei.com/vosk/models/model-list.json")
            for record in json_data:
                print(record.get('name') + " - " + record.get("obsolete") + " - " + record.get('url'))
                if record.get('name') == vosk_model_name and record.get('obsolete') == "false":
                    print('found!')
                    return record.get('url')
        except:
            return "https://alphacephei.com/vosk/models/vosk-model-small-en-us-0.15.zip"

    def set_speech_recognition_details(self, grammar_file: str, vosk_model: str, required_confidence: float = 80) -> bool:
        try:
            self.grammar_parser = GrammarParser()
            self.grammar_parser.set_grammar_file(grammar_file)

            if vosk_model is not None:
                if vosk_model.count("/") > 0 or vosk_model.count("\\") > 0:
                    logging.info("Setting VOSK model path as " + vosk_model)
                    self.model = Model(vosk_model)
                else:
                    logging.info("Downloading VOSK model " + vosk_model)
                    model_path = self.download_vosk_model(vosk_model)
                    logging.info("Setting VOSK model path as " + model_path)
                    self.model = Model(model_path)
            else:
                logging.info("Downloading default VOSK model vosk-model-small-en-us-0.15")
                model_path = self.download_vosk_model("vosk-model-small-en-us-0.15")
                logging.info("Setting VOSK model path as " + model_path)
                self.model = Model(model_path)

            self.required_confidence = required_confidence
            return True
        except Exception as e:
            logging.error("Unable to start speech recognition: " + repr(e))
            print("Unable to start speech recognition", str(e))
            return False

    async def start_speech_recognition(self, context):
        asyncio.create_task(self.process_recognition_queue())
        await asyncio.to_thread(self.listen)
        await asyncio.sleep(10)

    async def process_recognition_queue(self):
        logging.info("Started processing speech recognition queue")
        print("Started processing speech recognition queue")
        while not self.shutdown_event.is_set():  # Keep running unless shutdown is triggered
            try:
                item = await asyncio.wait_for(self.recognition_queue.get(), timeout=1.0)  # Wait for 1 second
                await self.process_speech(item)
            except asyncio.TimeoutError:
                continue  # Retry checking
        self.recognition_queue.task_done()
        logging.info("Stopped processing speech recognition queue")
        print("Stopped processing speech recognition queue")

    def listen(self):
        if self.stop_speech_recognition_event:
            self.stop_speech_recognition_event.set()

        stop_speech_recognition_event = asyncio.Event()
        self.stop_speech_recognition_event = stop_speech_recognition_event

        device_info = sd.query_devices(sd.default.device[0], 'input')
        samplerate = int(device_info['default_samplerate'])
        speech_queue = queue.Queue()

        def record_callback(indata, frames, time, status):
            if status:
                logging.error(str(status))
            speech_queue.put(bytes(indata))

        words_json = json.dumps(self.grammar_parser.all_words)
        recognizer = KaldiRecognizer(self.model, samplerate, words_json)
        recognizer.SetWords(False)
        recognizer.SetGrammar(words_json)

        try:
            logging.info("Started listening to voice via VOSK")
            print("Started listening to voice via VOSK")
            with sd.RawInputStream(dtype='int16',
                                   channels=1,
                                   callback=record_callback):
                response = speech_service_pb2.SpeechServiceResponse()
                response.speech_recognition_started.successful = True
                if self.grpc_response_queue:
                    asyncio.run(self.grpc_response_queue.put(response))
                while not stop_speech_recognition_event.is_set() and not self.shutdown_event.is_set():
                    data = speech_queue.get()
                    if recognizer.AcceptWaveform(data):
                        recognized_text = recognizer.Result()
                        if recognized_text:
                            asyncio.run(self.recognition_queue.put(recognized_text))
            logging.info("Stopped listening to voice via VOSK")
            print("Stopped listening to voice via VOSK")
        except KeyboardInterrupt:
            print('Finished recording due to keyboard interrupt')
            logging.error("Finished recording due to keyboard interrupt")
        except Exception as e:
            print("Error from VOSK: " + str(e))
            logging.error("Error from VOSK: " + str(e))
            logging.error(e)
            logging.error(traceback.format_exc())

    async def process_speech(self, recognizer_result: str):
        try:
            result_dict = json.loads(recognizer_result)
            text_recognized = result_dict.get("text", "")
            if not text_recognized == "":

                match = self.grammar_parser.find_match(text_recognized, self.required_confidence)

                if match is not None:
                    logging.info("Matched text \"" + match.matched_text + "\" (heard \"" + text_recognized + "\"")
                    if self.grpc_response_queue:
                        response = speech_service_pb2.SpeechServiceResponse()
                        response.speech_recognized.heard_text = text_recognized
                        response.speech_recognized.recognized_text = match.matched_text
                        response.speech_recognized.recognized_rule = match.rule
                        response.speech_recognized.confidence = round(match.confidence, 2)
                        response.speech_recognized.semantics.update(match.values)
                        await self.grpc_response_queue.put(response)
                    else:
                        print("I heard: '" + text_recognized + "', but I am " + str(round(match.confidence, 2)) + "% sure you said '" + match.matched_text + "'")
                else:
                    logging.debug("Recognized text " + text_recognized)

        except Exception as e:
            print("Error processing speech: " + str(e))
            logging.error("Error processing speech")
            logging.error(e)
            logging.error(traceback.format_exc())

    def set_grpc_response_queue(self, queue: asyncio.Queue):
        self.grpc_response_queue = queue

    def stop_speech_recognition(self):
        if self.stop_speech_recognition_event is not None:
            self.stop_speech_recognition_event.set()

    def shutdown(self):
        self.shutdown_event.set()



import subprocess
from pathlib import Path
from typing import Optional

import yapper.constants as c
from platformdirs import user_data_dir
from yapper import PiperSpeaker, PiperVoiceUS, PiperVoiceUK
from yapper.utils import (
    PLATFORM,
)

from py_speech_service.downloader import download_piper, download_piper_model


class Piper(PiperSpeaker):

    voice_onnx_files: dict[PiperVoiceUS | PiperVoiceUK, str] = {}
    voice_conf_files: dict[PiperVoiceUS | PiperVoiceUK, str] = {}

    def __init__(self, onnx_path: Optional[str] = None, conf_path: Optional[str] = None, piper_voice: str = "", alt_piper_voice: str = ""):
        app_dir = Path(user_data_dir("py_speech_service"))
        download_piper(app_dir)
        self.exe_path = str(
            app_dir
            / "piper"
            / ("piper.exe" if PLATFORM == c.PLATFORM_WINDOWS else "piper")
        )

        if onnx_path and conf_path:
            self.onnx_f = onnx_path
            self.conf_f = conf_path
        else:
            voice = self.string_to_voice(piper_voice)
            quality = PiperSpeaker.VOICE_QUALITY_MAP[voice]
            onnx_f, conf_f = download_piper_model(
                voice, quality, app_dir
            )
            self.onnx_f, self.conf_f = str(onnx_f), str(conf_f)
            self.voice_onnx_files[voice] = self.onnx_f
            self.voice_conf_files[voice] = self.conf_f

            if alt_piper_voice and alt_piper_voice != piper_voice:
                voice = self.string_to_voice(alt_piper_voice)
                quality = PiperSpeaker.VOICE_QUALITY_MAP[voice]
                onnx_f, conf_f = download_piper_model(
                    voice, quality, app_dir
                )
                onnx_f, conf_f = str(onnx_f), str(conf_f)
                self.voice_onnx_files[voice] = onnx_f
                self.voice_conf_files[voice] = conf_f

    @staticmethod
    def string_to_voice(voice: str) -> PiperVoiceUS | PiperVoiceUK:
        try:
            return PiperVoiceUS(voice)
        except ValueError:
            try:
                return PiperVoiceUK(voice)
            except ValueError:
                return PiperVoiceUS.HFC_FEMALE

    def set_speech_settings(self, onnx_path: Optional[str] = None, conf_path: Optional[str] = None, piper_voice: str = ""):
        app_dir = Path(user_data_dir("py_speech_service"))
        if onnx_path and conf_path:
            self.onnx_f = onnx_path
            self.conf_f = conf_path
        else:
            voice = self.string_to_voice(piper_voice)
            if self.voice_onnx_files.__contains__(voice):
                self.onnx_f = self.voice_onnx_files[voice]
                self.conf_f = self.voice_conf_files[voice]
            else:
                quality = PiperSpeaker.VOICE_QUALITY_MAP[voice]
                self.onnx_f, self.conf_f = download_piper_model(
                    voice, quality, app_dir
                )
                self.onnx_f, self.conf_f = str(self.onnx_f), str(self.conf_f)
                self.voice_onnx_files[voice] = self.onnx_f
                self.voice_conf_files[voice] = self.conf_f

    def text_to_wav(self, text: str, file: str, rate: float = 1):
        length_scale = 1 / rate
        subprocess.run(
            [
                self.exe_path,
                "-m",
                self.onnx_f,
                "-c",
                self.conf_f,
                "-f",
                file,
                "-q",
                "--length_scale",
                str(length_scale)
            ],
            input=text.encode("utf-8"),
            stdout=subprocess.DEVNULL,
            check=True
        )
